import json
import threading
import time

import click
from bottle import route, run, request
from pyngrok import ngrok

from sns_bridge.chalice import trigger_chalice_method
from sns_bridge.sns import subscribe_to_sns, unsubscribe_from_sns, confirm_sns
from sns_bridge.wsgi import ThreadedWSGIServer

subscription_arn = None
sns_topic_name = None


@route('/', method='POST')
def index():
    global subscription_arn
    body = json.loads(request.body.read())
    if body:
        if body.get('Type') == 'SubscriptionConfirmation':
            subscription_arn = confirm_sns(body.get('SubscribeURL'))
            print("Ready to receive SNS messages!")
            return "OK"
        if body.get('Type') == 'Notification':
            # Call the method!
            trigger_chalice_method(sns_topic_name, subscription_arn, body)
            return "OK"

    return "Hello"


@click.command()
@click.option('--role-arn', help='ARN of the role to assume')
@click.option('--port', help='Port number to run', default=65500)
@click.option('--topic-name', help='SNS topic name', required=True)
def main(topic_name, role_arn, port):
    global subscription_arn, sns_topic_name
    sns_topic_name = topic_name

    # Start web server to receive SNS messages
    threaded_wsgi = ThreadedWSGIServer(port=port)
    http_server = threading.Thread(target=run, kwargs={'server': threaded_wsgi})
    http_server.start()

    # Start local tunnel to connect to SNS
    http_tunnel = ngrok.connect(addr=port)

    # Subscribe to SNS via the tunnel
    url = http_tunnel.public_url.replace('http:', 'https:')
    subscribe_to_sns(url, topic_name, role_arn)

    # Wait
    try:
        while http_server.is_alive():
            time.sleep(1)
    except (KeyboardInterrupt, SystemExit):
        print("Stop received! Stopping web server...")
        threaded_wsgi.stop()

    # After we're done, unsubscribe our subscription
    print("Unsubscribing from SNS topic...")
    unsubscribe_from_sns(subscription_arn)


# Start our web server on a background thread, then proceed.
if __name__ == "__main__":
    main()



from bottle import ServerAdapter


class ThreadedWSGIServer(ServerAdapter):
    srv = None

    def run(self, app): # pragma: no cover
        from wsgiref.simple_server import WSGIRequestHandler, WSGIServer
        from wsgiref.simple_server import make_server
        import socket

        class FixedHandler(WSGIRequestHandler):
            def address_string(self): # Prevent reverse DNS lookups please.
                return self.client_address[0]
            def log_request(*args, **kw):
                if not self.quiet:
                    return WSGIRequestHandler.log_request(*args, **kw)

        handler_cls = self.options.get('handler_class', FixedHandler)
        server_cls  = self.options.get('server_class', WSGIServer)

        if ':' in self.host: # Fix wsgiref for IPv6 addresses.
            if getattr(server_cls, 'address_family') == socket.AF_INET:
                class server_cls(server_cls):
                    address_family = socket.AF_INET6

        self.srv = make_server(self.host, self.port, app, server_cls, handler_cls)
        self.srv.serve_forever()

    def stop(self):
        self.srv.shutdown()


import boto3
import xml.etree.ElementTree as ET

import requests

client = boto3.client('sns', region_name='us-west-2')


def assume_role(role_arn):
    global client

    sts_client = boto3.client('sts', region_name='us-west-2')

    assumed_role_obj = sts_client.assume_role(
        RoleArn=role_arn,
        RoleSessionName='SNSBridge',
    )

    credentials = assumed_role_obj['Credentials']

    client = boto3.client('sns',
                          aws_access_key_id=credentials['AccessKeyId'],
                          aws_secret_access_key=credentials['SecretAccessKey'],
                          aws_session_token=credentials['SessionToken'],
                          region_name='us-west-2')


def subscribe_to_sns(url, sns_endpoint_name, role_arn=None):
    """
    Subscribe a given URL to an SNS endpoint
    :param url: URL to subscribe
    :param sns_endpoint_name: Endpoint name to subscribe to
    :param role_arn: role arn to assume
    :return:
    """
    # TODO: properly handle pagination
    if role_arn:
        assume_role(role_arn)

    topics = client.list_topics()['Topics']
    topic_arn = None

    for topic in topics:
        if sns_endpoint_name in topic['TopicArn']:
            topic_arn = topic['TopicArn']
            break

    results = client.subscribe(TopicArn=topic_arn, Protocol='https', Endpoint=url)

    return results['SubscriptionArn']


def unsubscribe_from_sns(subscription_arn):
    client.unsubscribe(SubscriptionArn=subscription_arn)


def confirm_sns(confirmation_url):
    confirmation = requests.get(confirmation_url)
    node = ET.fromstring(confirmation.content)
    return node[0][0].text



import ast
import subprocess
from typing import Any


class Visitor(ast.NodeVisitor):
    def __init__(self, sns_topic_name):
        self._sns_topic_name = sns_topic_name
        self.method_name = None
        self.node = None

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Any:
        """
        Find function definitions, look for the decorators,
        and find our sns topic name in the arguments list
        :param node:
        :return:
        """
        for l in node.decorator_list:
            for a in l.args:
                if a.value == self._sns_topic_name:
                    print(node.name)
                    self.method_name = node.name
                    self.node = node


def trigger_chalice_method(topic_name, subscription_arn, body):
    """
    Trigger the Chalice method.

    :param topic_name: SNS topic name
    :param subscription_arn: Real-life subscription ARN
    :param body: The body of the SNS message
    :return:
    """

    # First, determine the method name to call by using AST
    # to visit the method decorators and find the SNS topic name
    f = open('app.py', 'r')
    v = Visitor(topic_name)
    node = ast.parse(f.read())
    v.visit(node)
    f.close()

    # unrelated, we need to fake the SNS message
    event = {
        'Records': [
            {
                'EventSource': 'aws:sns',
                "EventVersion": "1.0",
                "EventSubscriptionArn": subscription_arn,
                "Sns": body,
            }
        ]
    }

    code = f"""
from app import {v.method_name}
event = {event}
{v.method_name}(event, None)"""

    # Now, run the method in a new Python interpreter each time
    # to lazily replicate hot reloading
    with subprocess.Popen(["python", "-"],
                          stdout=subprocess.PIPE,
                          stdin=subprocess.PIPE,
                          universal_newlines=True,
                          bufsize=1) as p:
        p.stdin.write(code)
        p.stdin.close()
        for line in p.stdout:
            print(line, end='')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sns_bridge']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.21.8,<2.0.0',
 'bottle>=0.12.19,<0.13.0',
 'click>=8.0.4,<9.0.0',
 'pyngrok>=5.1.0,<6.0.0',
 'requests>=2.27.1,<3.0.0']

entry_points = \
{'console_scripts': ['sns_bridge = sns_bridge.main:main']}

setup_kwargs = {
    'name': 'sns-bridge',
    'version': '0.1.1',
    'description': 'Run an AWS Chalice SNS listener locally',
    'long_description': None,
    'author': 'Jake Wood',
    'author_email': 'jake@testbox.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
